using System;
using System.IO;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using MeshVR;
using MVR.FileManagementSecure;
using AssetBundles;
using SimpleJSON;

namespace Stopper {
	public class AlternativeFuta : MVRScript {
	
	// Character skin prefabs
	GameObject femaleSkinsPrefab;
	GameObject maleSkinsPrefab;
	
	GameObject femaleSkinsContainer;
	
	// Stuff for the new 'DAZCharacter'
	// that has our new modified skin
	GameObject newCharacterContainer;
	DAZCharacter newCharacter;
	
	DAZCharacterSelector characterSelector;
	
	DAZCharacterRun characterRun;
	
	DAZMergedSkinV2 skin;
		
	DAZMergedMesh mergedMesh;
	
	DAZMesh genMesh;
	
	GameObject femaleSkinBase;
	
	DAZBone gen1Bone;
	DAZBone gen2Bone;
	DAZBone gen3Bone;
	
	// We also instantiate a male character to steal
	// some data from during initialization
	GameObject maleCharacterContainer;
	GameObject maleSkinBase;
	DAZMergedSkinV2 maleSkin;
	DAZMergedMesh maleMergedMesh;
	
	// UI Stuff
	GameObject UIContainer; 
	Material genMaterial;
	int genMaterialIndex;
	MaterialOptionsWrapper genMaterialOptions;
	
	GameObject materialUIPrefab;
	MVRScriptUI scriptUI;
	
	
	// Morph Stuff
	float genBoneYOffsetDefault = 0.028f;
	float genBoneZOffsetDefault = -0.016f;
	
	DAZMorph genYOffsetMorph;
	DAZMorph genZOffsetMorph;
	
	DAZMorph genMeshYOffsetMorph;
	DAZMorph genMeshZOffsetMorph;
	
	JSONClass savedSceneConfig;
	
	// Physics stuff
	List<string> genAutoColliderIgnoreNames = new List<string>{
			// Front pelvis
			"AutoColliderpelvisF1","AutoColliderpelvisF2","AutoColliderpelvisF3",
			"AutoColliderpelvisF4","AutoColliderpelvisF5","AutoColliderpelvisF6",
			// Inner pelvis
			"AutoColliderpelvisFL1","AutoColliderpelvisFL2","AutoColliderpelvisFL3","AutoColliderpelvisFL4",
			"AutoColliderpelvisFR1","AutoColliderpelvisFR2","AutoColliderpelvisFR3","AutoColliderpelvisFR4"
	};
	List<AutoCollider> genAutoColliderIgnoreList = new List<AutoCollider>();
	
	// Small wrappers that help us get access to some protected members
	public class DAZSkinV2Wrapper : DAZSkinV2 {
		public void Copy(DAZSkinV2 other){
			this.sceneGeometryId = other.sceneGeometryId;
			this._numBones = other.numBones;
			this.nodes = other.nodes;
		}
	}

	public class MaterialOptionsWrapper : DAZCharacterMaterialOptions {
		public void SetParams(){
			this.SetAllParameters();
		}
	}

	public override void Init() {
		
		// Initialize some references. 
		// TODO: Replace these with hard-coded paths
		foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
			if(go.name =="FemaleCharactersPrefab"){
				femaleSkinsPrefab = go;
			}
			
			if(go.name =="MaleCharactersPrefab"){
				maleSkinsPrefab = go;
			}
		}
		
		foreach(GameObject go in GetAllChildren(containingAtom.gameObject)){
			if(go.name.Contains("FemaleCharactersPrefab")){
				femaleSkinsContainer = go;
			}
		}
		
		foreach(GameObject go in GetAllChildren(femaleSkinsPrefab)){
			if(go.name == "rensimone"){
				femaleSkinBase = go;
			}
		}
		
		foreach(GameObject go in GetAllChildren(maleSkinsPrefab)){
			if(go.name == "FutaSimone"){
				maleSkinBase = go;
			}
		}
		
		foreach(AutoCollider ac in containingAtom.GetComponentsInChildren<AutoCollider>()){
			if(genAutoColliderIgnoreNames.Contains(ac.name)){
				genAutoColliderIgnoreList.Add(ac);
			}
		}
		
		characterRun = containingAtom.GetComponentInChildren<DAZCharacterRun>();
		
		gen1Bone = characterRun.bones.GetDAZBone("Gen1");
		gen2Bone = characterRun.bones.GetDAZBone("Gen2");
		gen3Bone = characterRun.bones.GetDAZBone("Gen3");
		
		characterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
	}
	
	public void Start() {
		StartCoroutine("TransPlant");
	}
	
	private IEnumerator TransPlant(){
		
		
		/* 	The VAM skin hierarchy includes a DAZMergedSkinV2, which uses a DAZMergedMesh. 
			'Merged' here refers to the fact that genitals have their own DAZMesh that is 
			grafted onto the main mesh. There is code for 2 graft meshes and, luckily, the 
			female model only uses one. 
		
			The first thing we'll do is set up a new skin. This requires a few high-level steps:
			1. Create a new DAZCharacter off of a prefab
				- This will create most of our DAZMergedMesh and our DAZMergedSkinV2
			2. Add a DAZMesh and DAZSkinV2 for the penis prior to the main skin initializing
				- The DAZMesh we get from a file
				- The DAZSkinV2 we'll copy from a male skin
			3. Merge the DAZMergedMesh and our DAZMergedSkinV2
			4. Set the materials
		*/
		
		// Initialize the male skin that we'll steal some data from.
		// Specifically, we want the node weights, which determine how 
		// each vertex is weiged to the DAZBones of the skeleton
		maleCharacterContainer = UnityEngine.Object.Instantiate<GameObject>(maleSkinBase);
		
		// Enable the object. This will cause the DAZCharacter to load skin
		// artifacts from assetBundles
		maleCharacterContainer.SetActive(true);
		
		DAZMergedSkinV2 maleMergedSkin = null;
		while(maleMergedSkin == null){
			maleMergedSkin = maleCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();
			yield return null;
		}
		maleCharacterContainer.SetActive(false);
		maleMergedMesh = (maleMergedSkin.dazMesh as DAZMergedMesh);
			
	
		// Initialize the new skin, which will initially include just the DAZCharacter
		newCharacterContainer = UnityEngine.Object.Instantiate<GameObject>(femaleSkinBase);
		newCharacterContainer.SetActive(false);
		newCharacterContainer.name = "FutaSkin";
		
		newCharacter = newCharacterContainer.GetComponentInChildren<DAZCharacter>();
		newCharacterContainer.transform.parent = femaleSkinsContainer.transform;
	
		newCharacter.rootBonesForSkinning = characterSelector.rootBones;
		
		// Once a DAZSkinV2 is initialized, all the vertices are copied into ComputeBuffers
		// and I don't know how to reset those. This means that we need to do all of our mesh
		// modification prior to initialization.
		
		// Initialization happens in the onLoadedHandler, so in order to wedge ourselves in 
		// between the DAZCharacter loading the assets and initialization, we'll just also
		// add ourselves to the onLoadedHandler (which DAZCharacter will later combine with its own)
		newCharacter.onLoadedHandlers = () => {
			
			// Skin should be loaded (but not initialized) by this point
			skin = newCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();	
			mergedMesh = (skin.dazMesh as DAZMergedMesh);
			
			
			// Load our mesh from file. This is the only way I found to initialize the
			// vertices manually. 
			// The file I'm using came from using StoreToBinaryFile, which required me to 
			// disable file IO restrictions. (Luckily LoadFromBinaryFile doesn't have those)
			genMesh = mergedMesh.gameObject.AddComponent<DAZMesh>();
			genMesh.LoadFromBinaryFile($"{GetPluginPath(this,containingAtom)}/male_gen_dazmesh");
			
			// The meshgraft includes instructions on which vertices should
			// be 'glued' to the base model. We're not bothering with this for now.
			// TODO: Set this...
			genMesh.meshGraft = new DAZMeshGraft();
			genMesh.meshGraft.vertexPairs = new DAZMeshGraftVertexPair[0];
			genMesh.meshGraft.hiddenPolys = new int[0];
			genMesh.graftTo = mergedMesh.targetMesh;
			
			genMesh.Init();
		
			// Merge the skin! 
			// This should result in the mergedMesh copying all of our gen mesh's 
			// vertices. 
			mergedMesh.Merge();

			
			
			// DAZSkinV2 also has a merge concept, which copies over the bone weights
			// that determine how the mesh is skinned to the skeleton.
			
			// For this we'll use the male skin we created earlier and copy its verts
			DAZSkinV2 maleGenSkin = null;
			foreach(var skinv2 in maleCharacterContainer.GetComponentsInChildren<DAZSkinV2>()){
				if(skinv2.dazMesh == maleMergedMesh.graftMesh){
					maleGenSkin = skinv2;
				}
			}
			DAZSkinV2Wrapper genSkinv2 = mergedMesh.gameObject.AddComponent<DAZSkinV2Wrapper>();
			genSkinv2.Copy(maleGenSkin);
			
			// Merge the skins.
			skin.Merge();
	
			// Set the materials
			
			foreach(Material material in maleMergedSkin.GPUmaterials){
				if (material.name == "Genitalia-1") {
					genMaterial = material;
				}
			}
			
			Material[] originalMaterials = skin.GPUmaterials;
			skin.CopyMaterials();
			for (int i = 0; i < skin.dazMesh.materials.Length-1; i++)	{
				if(i < originalMaterials.Length){
					skin.GPUmaterials[i] = originalMaterials[i];
				} else if (skin.GPUmaterials[i].name == "Genitalia") {
					if(genMaterial == null){
						genMaterial = new Material(Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff"));
					}
					genMaterialIndex = i;
					skin.GPUmaterials[i] = genMaterial;
				} else {
					skin.GPUmaterials[i] = new Material(Shader.Find("Custom/Discard"));
				}
			}
			
			skin.FlushBuffers();
		}; // End onLoadedHandlers


		// Now that we have our load handler set up, enable the container so that
		// the DAZCharacter loads 
		newCharacterContainer.SetActive(true);

		// Wait for it do its thing...
		while(skin == null){
			skin = newCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();
			yield return null;
		}
		
		skin.name = "FutaSkin";
		newCharacterContainer.SetActive(false);
		
		
		// Set the new skin as the selected character
		
		SetMaleTransformsActive(false); // In case they're enabled for some reason
	
		characterSelector.selectedCharacter = newCharacter;
		
		// There are 3 morph banks for the (up to) 3 grafted meshes.
		// We want our 2nd graft (morphBank3) to use the morph's for 
		// The male's 1st graft (maleMorphBank2)
		characterRun.morphBank3 = characterSelector.maleMorphBank2;
		
		// morph banks will try to apply morphs to connected meshes
		// This might be set if the model was previously male. Need to unset it
		characterRun.morphBank3.connectedMesh = null;
		characterRun.ResetMorphs();	
		characterSelector.maleMorphBank2.gameObject.SetActive(true);
		
		// Set up Material Options 
		while(genMaterial == null){
			yield return null;
		}
		
		genMaterialOptions = newCharacterContainer.AddComponent<MaterialOptionsWrapper>();
		
		genMaterialOptions.textureGroup1 = new MaterialOptionTextureGroup();
		genMaterialOptions.SetCustomTextureFolder("Custom/Atom/Person/Textures");

		genMaterialOptions.paramMaterialSlots = new int[]{genMaterialIndex};
		genMaterialOptions.textureGroup1.materialSlots = genMaterialOptions.paramMaterialSlots;
		
		genMaterialOptions.skin = skin;
		genMaterialOptions.materialForDefaults = genMaterial;
		genMaterialOptions.enabled = true;
		
		// Set up the UI
		
		
		// The material selction prefab doesn't do well with a layout,
		// so we're just going to remove it.
		while(scriptUI == null){
			if(this.UITransform != null){
				scriptUI = this.UITransform.GetComponentInChildren<MVRScriptUI>();
			}
			yield return null;
		}
		
		if(scriptUI == null){
			SuperController.LogMessage("Unable to load script UI.");
			yield break;
		}
		
		UnityEngine.UI.VerticalLayoutGroup layout = scriptUI.fullWidthUIContent.gameObject.GetComponent<UnityEngine.UI.VerticalLayoutGroup>();
		Destroy(layout);
		
		// Find a pre-existing material option UI to use as a template. 
		// Any would likely do...
		while( materialUIPrefab == null){
			foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
				if(go.name =="Mouth Materials") {
					materialUIPrefab = go;
					break;
				}
			}
			if(materialUIPrefab == null){
				yield return null;
			}
		}
		
		var matOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
		genMaterialOptions.SetUI(matOptionsUI);
		
		genMaterialOptions.SyncAllParameters();
		
		genMaterialOptions.InitUI();
			
		if(savedSceneConfig != null){
			JSONClass matOptionsJson = savedSceneConfig["genMaterialOpts"] as JSONClass;
			
			if(matOptionsJson != null){
				genMaterialOptions.RestoreFromJSON(matOptionsJson);
				genMaterialOptions.SetParams();
			}
		}		

		matOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
		scriptUI.fullWidthUIContent.gameObject.SetActive(true);
		matOptionsUI.gameObject.SetActive(true);
		
		SetUpMorphs();
		
		// Re-enable the male portions of the model
		SetMaleTransformsActive(true);
	}
	
	private void SetUpMorphs(){
		var bank = characterSelector.maleMorphBank2;
		
		// Bone+Mesh Morphs
		genYOffsetMorph = bank.GetMorph("maleGenYOffset");
		if(genYOffsetMorph == null){
			genYOffsetMorph = new DAZMorph() {morphName = "maleGenYOffset", displayName = "AltFuta Bone+Mesh Y-Offset"};
			
			SetUpBoneMorph(genYOffsetMorph, DAZMorphFormulaTargetType.BoneCenterY, new Vector3(0f,1f,0f));
			
			bank.AddMorphUsingSubBanks(genYOffsetMorph);
		} else {
			SetUpBoneMorph(genYOffsetMorph, DAZMorphFormulaTargetType.BoneCenterY, new Vector3(0f,1f,0f));
			genYOffsetMorph.Init();
		}
		genYOffsetMorph.morphValue = genBoneYOffsetDefault;
		genZOffsetMorph = bank.GetMorph("maleGenZOffset");
		if(genZOffsetMorph == null){
			genZOffsetMorph = new DAZMorph() {morphName = "maleGenZOffset", displayName = "AltFuta Bone+Mesh Z-Offset"};
			
			SetUpBoneMorph(genZOffsetMorph, DAZMorphFormulaTargetType.BoneCenterZ, new Vector3(0f,0f,1f));
			
			bank.AddMorphUsingSubBanks(genZOffsetMorph);
		} else {
			SetUpBoneMorph(genZOffsetMorph, DAZMorphFormulaTargetType.BoneCenterZ, new Vector3(0f,0f,1f));
			genZOffsetMorph.Init();
		}
		genZOffsetMorph.morphValue = genBoneZOffsetDefault;
		
		// Mesh-only morphs
		genMeshYOffsetMorph = bank.GetMorph("maleGenMeshYOffset");
		if(genMeshYOffsetMorph == null){
			genMeshYOffsetMorph = new DAZMorph() {morphName = "maleGenMeshYOffset", displayName = "AltFuta Mesh Y-Offset"};
			SetUpMeshMorph(genMeshYOffsetMorph, new Vector3(0f,1f,0f));
			
			bank.AddMorphUsingSubBanks(genMeshYOffsetMorph);
		} else {
			SetUpMeshMorph(genMeshYOffsetMorph, new Vector3(0f,1f,0f));
			genMeshYOffsetMorph.Init();
		}
		genMeshYOffsetMorph.morphValue = 0f;
		
		genMeshZOffsetMorph = bank.GetMorph("maleGenMeshZOffset");
		if(genMeshZOffsetMorph == null){
			genMeshZOffsetMorph = new DAZMorph() {morphName = "maleGenMeshZOffset", displayName = "AltFuta Mesh Z-Offset"};
			SetUpMeshMorph(genMeshZOffsetMorph, new Vector3(0f,0f,1f));
			
			bank.AddMorphUsingSubBanks(genMeshZOffsetMorph);
		} else {
			SetUpMeshMorph(genMeshZOffsetMorph, new Vector3(0f,0f,1f));
			genMeshZOffsetMorph.Init();
		}
		genMeshZOffsetMorph.morphValue = 0f;
		
		
		characterSelector.RefreshRuntimeMorphs();
	}

	private void SetUpMeshMorph(DAZMorph morph, Vector3 direction){	
		morph.region = "Genitals";
		morph.min = -0.1f;
		morph.max = 0.1f;
		morph.visible = true;
		
		// Vertex Deltas
		var deltas = new List<DAZMorphVertex>();
		for(int i = 0; i < genMesh.baseVertices.Length; i++){
			deltas.Add(new DAZMorphVertex(){vertex = i, delta = direction});
		}
		morph.deltas = deltas.ToArray();	
		morph.numDeltas = morph.deltas.Length;
		
		morph.formulas = new DAZMorphFormula[0];
	}
	private void SetUpBoneMorph(DAZMorph morph, DAZMorphFormulaTargetType targetType, Vector3 direction){
		SetUpMeshMorph(morph,direction);

		// Bone Morphs
		var boneFormulas = new List<DAZMorphFormula>();
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen1", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen2", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen3", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Testes", multiplier = 1f, targetType = targetType});
		
		morph.formulas = boneFormulas.ToArray();
	}
	
	private DAZMorphFormula MorphBoneFormula(string boneName, DAZMorphFormulaTargetType targetType){
		var formula = new DAZMorphFormula();
		formula.targetType = targetType;
		formula.target = boneName;
		formula.multiplier = 1f;
		return formula;
	}
	
	private string GetPathToObject(GameObject go){
		 if (go.transform.parent == null)
			 return "/" + go.name;
		 return GetPathToObject(go.transform.parent.gameObject) + "/" + go.name;
	}

	/* Change the status of the male portions of the model.
		Specifically we only want the rigid bodies and controllers
	*/
	private void SetMaleTransformsActive(bool active){
		// TODO: Use hardcoded paths for these
		foreach(GameObject go in GetAllChildren(containingAtom.gameObject)){
			if((go.name.Contains("penis") || go.name.Contains("testes")) 
				&& go.name.Contains("Control")){
				go.SetActive(active);
			}
			if(	go.name == "Gen1"
				|| go.name == "Gen2"
				|| go.name == "Gen3"
				|| go.name == "Testes"){
				go.SetActive(active);
				SyncAutoColliderIgnore(go);
			}
		}
	}
	
	private void SyncAutoColliderIgnore(GameObject go){
		if(go.activeInHierarchy){
			foreach(AutoCollider ac in genAutoColliderIgnoreList){
				var ignoreList = ac.ignoreColliders.ToList();
				if(!ignoreList.Contains(go.transform)){
					ignoreList.Add(go.transform);
					ac.ignoreColliders = ignoreList.ToArray();
					ac.InitColliders();
				}
			}
		}
	}
	
	private void Update(){
		if(characterRun.morphBank3 != null){
			// Fix these settings in case they got changed somehow
			if(characterRun.morphBank3 != characterSelector.maleMorphBank2){
				characterRun.morphBank3 = characterSelector.maleMorphBank2;
			}

			if(characterRun.morphBank3.connectedMesh != null){
				characterRun.morphBank3.connectedMesh = null;
			}
			
			// For whatever reason, the threaded morph logic in DAZCharacterRun
			// just doesn't bother with the 2nd graft. 
			// So we'll just do it manually here
			if ( characterRun.morphBank3.bonesDirty)	{
				characterRun.bones.SetMorphedTransform(false);
				characterRun.morphBank3.bonesDirty = false;
			}
			if (characterRun.morphBank3.boneRotationsDirty != null)
			{
				foreach (var entry in characterRun.morphBank3.boneRotationsDirty){
					entry.Key.SyncMorphBoneRotations(false);
				}
				characterRun.morphBank3.boneRotationsDirty.Clear();
			}
		}
	}
	
	private void OnDestroy()
	{	
		if(maleCharacterContainer != null){
			DestroyImmediate(maleCharacterContainer);
		}
		if(newCharacterContainer != null){
			DestroyImmediate(newCharacterContainer);
		}
	}
		
	List<GameObject> GetAllChildren(GameObject parent, List<GameObject> objectList = null)
	{
		if (objectList == null) objectList = new List<GameObject>();
	  
		foreach (Transform child in parent.transform) {
			objectList.Add(child.gameObject);
			GetAllChildren(child.gameObject, objectList);
		}
		return objectList;
	}
	
	public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
		savedSceneConfig = jc;
		
		base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
	}
	
	public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
	{
		var jsonClass = base.GetJSON(true, true, true);
		
		jsonClass.Add("genMaterialOpts", genMaterialOptions.GetJSON(true,true,true)); 
		return jsonClass;
	}
	
	// Function taken from VAMDeluxe's code
	public static string GetPluginPath(MVRScript plugin, Atom atom){
		SuperController.singleton.currentSaveDir = SuperController.singleton.currentLoadDir;
		string pluginId = plugin.storeId.Split('_')[0];
		MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
		string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
		string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }));
		return pathToScriptFolder;
	}
}
}